<?php

namespace Drupal\monarch_data_entity\Plugin\Field\FieldWidget;

use Drupal\Core\Field\FieldDefinitionInterface;
use Drupal\Core\Field\FieldItemListInterface;
use Drupal\Core\Field\WidgetBase;
use Drupal\Core\Form\FormStateInterface;
use Drupal\views\Entity\View;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Defines the 'monarch_data_entity_view' field widget.
 *
 * @FieldWidget(
 *   id = "monarch_data_entity_view",
 *   label = @Translation("View"),
 *   multiple_values = TRUE,
 *   field_types = {
 *     "entity_reference",
 *   },
 * )
 */
class ViewWidget extends WidgetBase {

  /**
   * The entity_type.manager service.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * The entity_field.manager service.
   *
   * @var \Drupal\Core\Entity\EntityFieldManagerInterface
   */
  protected $entityFieldManager;

  /**
   * The renderer service.
   *
   * @var \Drupal\Core\Render\RendererInterface
   */
  protected $renderer;

  /**
   * The token service.
   *
   * @var \Drupal\monarch_data_entity\TokenService
   */
  protected $tokenService;

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    $interface = parent::create($container, $configuration, $plugin_id, $plugin_definition);

    $interface->entityTypeManager = $container->get('entity_type.manager');
    $interface->entityFieldManager = $container->get('entity_field.manager');
    $interface->renderer = $container->get('renderer');
    $interface->tokenService = $container->get('monarch_data_entity.token_service');

    return $interface;
  }

  /**
   * {@inheritdoc}
   */
  public static function defaultSettings() {
    return parent::defaultSettings() + [
      'view' => NULL,
      'args' => [],
    ];
  }

  /**
   * {@inheritdoc}
   */
  public function settingsForm(array $form, FormStateInterface $form_state) {
    $options = [];
    $build = parent::settingsForm($form, $form_state) + [
      '#prefix' => '<div id="extra-options-wrapper">',
      '#suffix' => '</div>',
    ];
    $views = View::loadMultiple();
    $base_table = $this->entityTypeManager->getStorage($this->fieldDefinition->getFieldStorageDefinition()->getSetting('target_type'))->getBaseTable();

    foreach ($views as $view_id => $view) {
      foreach (($view->get('display') ?? []) as $display_id => $display_options) {
        if ($display_options['display_plugin'] == 'embed' && $view->get('base_table') == $base_table) {
          $options[$view->get('label')][$view_id . '.' . $display_id] = $display_options['display_title'] . ' (' . $view_id . '.' . $display_id . ')';
        }
      }
    }

    $ajax = [
      'callback' => [$this, 'ajaxSettings'],
      'disable-refocus' => FALSE,
      'event' => 'change',
      'wrapper' => 'extra-options-wrapper',
      'progress' => [
        'type' => 'throbber',
      ],
    ];

    $values = $form_state->getValues()['fields'][$this->fieldDefinition->get('field_name')]['settings_edit_form']['settings'] ?? $this->settings ?? NULL;

    $build['view'] = [
      '#type' => 'select',
      '#title' => $this->t('View'),
      '#options' => $options,
      '#empty_option' => $this->t('-- None --'),
      '#empty_value' => '',
      '#default_value' => $this->settings['view'] ?? NULL,
      '#ajax' => $ajax,
    ];

    if ($values['view'] ?? NULL) {
      [$selected_view_id, $selected_display_id] = explode('.', $values['view']);

      if ($selected_view = $views[$selected_view_id] ?? NULL) {
        $args = $this->getArgs($selected_view, $selected_display_id);

        if (!empty($args)) {
          $build['args']['#tree'] = TRUE;
          foreach ($args as $arg_name => $arg_settings) {
            $build['args'][$arg_name]['value'] = [
              '#type' => 'textfield',
              '#title' => $this->t('@arg_name value', ['@arg_name' => $arg_name]),
              '#default_value' => $values['args'][$arg_name]['value'] ?? NULL,
            ];

            $build['args'][$arg_name]['transformer'] = [
              '#type' => 'select',
              '#title' => $this->t('@arg_name type', ['@arg_name' => $arg_name]),
              '#options' => $this->transformTypeOptions(),
              '#empty_value' => NULL,
              '#empty_option' => $this->t('Literal'),
              '#default_value' => $values['args'][$arg_name]['transformer'] ?? NULL,
            ];
          }
        }
      }
    }

    $build['view_arguments'] = [];

    return $build;
  }

  /**
   * Ajax callback.
   */
  public function ajaxSettings(array &$form, FormStateInterface $form_state) {
    return $form['fields'][$this->fieldDefinition->get('field_name')]['plugin']['settings_edit_form']['settings'];
  }

  /**
   * {@inheritdoc}
   */
  public function settingsSummary() {
    return ['View: ' . ($this->settings['view'] ?? '')];
  }

  /**
   * Get a list of transform types.
   *
   * @return array
   *   A list of transform options.
   */
  public function transformTypeOptions() {
    return [
      'token' => $this->t('Query Token'),
    ];
  }

  /**
   * Get the view chosen in the options.
   *
   * @return \Drupal\views\Entity\View
   *   The view chosen in the options.
   */
  public function getView() : ?View {
    if ($this->settings['view'] ?? NULL ?: NULL) {
      [$view_id, $display_id] = explode('.', $this->settings['view'], 2);

      if ($view = View::load($view_id)) {
        return $view;
      }
    }

    return NULL;
  }

  /**
   * Get the display ID chosen in the options.
   *
   * @return string
   *   The display ID chosen in the options.
   */
  public function getDisplayId() : ?string {
    if ($this->settings['view'] ?? NULL ?: NULL) {
      [$view_id, $display_id] = explode('.', $this->settings['view'], 2);
      return $display_id;
    }

    return NULL;
  }

  /**
   * Get view arguments.
   *
   * @return array
   *   The argument list.
   */
  public function getArgs() {
    if ($view = $this->getView()) {
      $default_display = $view->getDisplay('default');
      $display_id = $this->getDisplayId();
      $display = $view->getDisplay($display_id);
      $args = ($display['display_options']['defaults']['arguments'] ?? TRUE) ? $default_display['display_options']['arguments'] : $display['display_options']['arguments'];
      return $args ?? [];
    }

    return [];
  }

  /**
   * {@inheritdoc}
   */
  public function formElement(FieldItemListInterface $items, $delta, array $element, array &$form, FormStateInterface $form_state) {
    $ids = [];

    $storage = $this->fieldDefinition->getFieldStorageDefinition();

    foreach ($items as $d => $item) {
      $ids[$d] = $item->target_id ?? '';
    }

    $element['target_id'] = [
      '#type' => 'view_input',
      '#cardinality' => $storage->getCardinality(),
      '#default_value' => implode(',', $ids),
    ];

    if ($view = $this->getView()) {
      $element['target_id']['#view_name'] = $view->id();
      $element['target_id']['#view_display_id'] = $this->getDisplayId();
      $element['target_id']['#view_args'] = [];
      $element['#process'][] = [$this, 'processElement'];
      $element['#after_build'][] = [$this, 'afterBuildElement'];
    }

    return $element;
  }

  /**
   * The '#process' callback.
   */
  public function processElement(array $element, FormStateInterface $form_state) {
    $form = $form_state->getCompleteForm();

    foreach ($this->getArgs() as $arg_id => $arg_config) {
      if (isset($this->settings['args'][$arg_id]['value'])) {
        if (
          ($value_transformer = $this->settings['args'][$arg_id]['transformer'] ?? NULL) &&
          $value_transformer === 'token'
        ) {
          $this_field_name = $this->fieldDefinition->getName();

          $form[$this_field_name]['#prefix'] = "<div id=\"$this_field_name-wrapper\">";
          $form[$this_field_name]['#suffix'] = "</div>";

          $element['target_id']['#ajax'] = [
            'callback' => [$this, 'ajax'],
            'disable-refocus' => TRUE,
            'event' => 'change',
            'wrapper' => "$this_field_name-wrapper",
            'element' => $this_field_name,
            'progress' => [
              'type' => 'throbber',
            ],
          ];
        }
      }
    }

    return $element;
  }

  /**
   * The '#after_build' callback.
   */
  public function afterBuildElement(array $element, FormStateInterface $form_state) {
    $form = $form_state->getCompleteForm();

    foreach ($this->getArgs() as $arg_id => $arg_config) {
      if (isset($this->settings['args'][$arg_id]['value'])) {
        if ($value_transformer = $this->settings['args'][$arg_id]['transformer'] ?? NULL) {
          $element['target_id']['#view_args'][] = $this->transformArgumentValue($this->settings['args'][$arg_id]['value'], $value_transformer, $element['target_id'], $arg_config, $form, $form_state);
        }
        else {
          $element['target_id']['#view_args'][] = $this->settings['args'][$arg_id]['value'];
        }
      }
    }

    return $element;
  }

  /**
   * Transforms argument values according to the settings.
   *
   * @param string|null $value
   *   The transformation value.
   * @param string|null $transformer
   *   The transformation type.
   * @param array $element
   *   The element to transform the value for.
   * @param string|null $argument_settings
   *   The settings for this argument.
   * @param array $form
   *   The form containing the element.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The state of the form.
   *
   * @return string|null
   *   The transformed value.
   */
  public function transformArgumentValue(?string $value, ?string $transformer, array &$element, ?array $argument_settings, array &$form, FormStateInterface $form_state) : ?string {
    if ($transformer === 'token') {
      $ret = $this->tokenService->transformArgumentToken($value, $element, $argument_settings, $form, $form_state);

      return $ret;
    }

    return empty($transformer) ? ($value ?: NULL) : NULL;
  }

  /**
   * Ajax callback.
   */
  public function ajax(array &$form, FormStateInterface $form_state) {
    return $form[$form_state->getTriggeringElement()['#ajax']['element']];
  }

  /**
   * {@inheritdoc}
   */
  public function massageFormValues(array $values, array $form, FormStateInterface $form_state) {
    $ids = array_filter(explode(',', $values['target_id'] ?? ''));

    unset($values['target_id']);

    foreach ($ids as $d => $id) {
      $values[$d]['target_id'] = $id;
    }

    return $values;
  }

  /**
   * {@inheritdoc}
   */
  public static function isApplicable(FieldDefinitionInterface $field_definition) {
    return TRUE;
  }

}
